@@ -1,7 +1,8 @@
 from abc import ABC, abstractmethod
-from collections.abc import MutableMapping
+from collections.abc import Awaitable, Callable, MutableMapping
 from contextlib import suppress
-from dataclasses import dataclass
+from dataclasses import dataclass, field
+from functools import partial
 from typing import (
     Any,
     AsyncContextManager,
@@ -13,8 +14,17 @@ from typing import (
 )
 
 from injection._core.common.asynchronous import Caller
-from injection._core.scope import Scope, get_active_scopes, get_scope
-from injection.exceptions import InjectionError
+from injection._core.common.asynchronous import (
+    create_semaphore as _create_async_semaphore,
+)
+from injection._core.scope import (
+    Scope,
+    get_scope,
+    in_scope_cache,
+    remove_scoped_values,
+)
+from injection._core.slots import SlotKey
+from injection.exceptions import EmptySlotError, InjectionError
 
 
 @runtime_checkable
@@ -37,14 +47,10 @@ class Injectable[T](Protocol):
         raise NotImplementedError
 
 
-@dataclass(repr=False, frozen=True, slots=True)
-class BaseInjectable[T](Injectable[T], ABC):
+@dataclass(repr=False, eq=False, frozen=True, slots=True)
+class SimpleInjectable[T](Injectable[T]):
     factory: Caller[..., T]
 
-
-class SimpleInjectable[T](BaseInjectable[T]):
-    __slots__ = ()
-
     async def aget_instance(self) -> T:
         return await self.factory.acall()
 
@@ -52,51 +58,78 @@ class SimpleInjectable[T](BaseInjectable[T]):
         return self.factory.call()
 
 
-class SingletonInjectable[T](BaseInjectable[T]):
-    __slots__ = ("__dict__",)
+class CacheLogic[T]:
+    __slots__ = ("__semaphore",)
+
+    __semaphore: AsyncContextManager[Any]
+
+    def __init__(self) -> None:
+        self.__semaphore = _create_async_semaphore(1)
+
+    async def aget_or_create[K](
+        self,
+        cache: MutableMapping[K, T],
+        key: K,
+        factory: Callable[..., Awaitable[T]],
+    ) -> T:
+        async with self.__semaphore:
+            with suppress(KeyError):
+                return cache[key]
+
+            instance = await factory()
+            cache[key] = instance
+
+        return instance
+
+    def get_or_create[K](
+        self,
+        cache: MutableMapping[K, T],
+        key: K,
+        factory: Callable[..., T],
+    ) -> T:
+        with suppress(KeyError):
+            return cache[key]
+
+        instance = factory()
+        cache[key] = instance
+        return instance
+
+
+@dataclass(repr=False, eq=False, frozen=True, slots=True)
+class SingletonInjectable[T](Injectable[T]):
+    factory: Caller[..., T]
+    cache: MutableMapping[str, T] = field(default_factory=dict)
+    logic: CacheLogic[T] = field(default_factory=CacheLogic)
 
     __key: ClassVar[str] = "$instance"
 
     @property
     def is_locked(self) -> bool:
-        return self.__key in self.__cache
-
-    @property
-    def __cache(self) -> MutableMapping[str, Any]:
-        return self.__dict__
+        return self.__key in self.cache
 
     async def aget_instance(self) -> T:
-        cache = self.__cache
-
-        with suppress(KeyError):
-            return cache[self.__key]
-
-        instance = await self.factory.acall()
-        cache[self.__key] = instance
-        return instance
+        return await self.logic.aget_or_create(
+            self.cache,
+            self.__key,
+            self.factory.acall,
+        )
 
     def get_instance(self) -> T:
-        cache = self.__cache
-
-        with suppress(KeyError):
-            return cache[self.__key]
-
-        instance = self.factory.call()
-        cache[self.__key] = instance
-        return instance
+        return self.logic.get_or_create(self.cache, self.__key, self.factory.call)
 
     def unlock(self) -> None:
-        self.__cache.pop(self.__key, None)
+        self.cache.pop(self.__key, None)
 
 
 @dataclass(repr=False, eq=False, frozen=True, slots=True)
 class ScopedInjectable[R, T](Injectable[T], ABC):
     factory: Caller[..., R]
     scope_name: str
+    logic: CacheLogic[T] = field(default_factory=CacheLogic)
 
     @property
     def is_locked(self) -> bool:
-        return any(self in scope.cache for scope in get_active_scopes(self.scope_name))
+        return in_scope_cache(self, self.scope_name)
 
     @abstractmethod
     async def abuild(self, scope: Scope) -> T:
@@ -107,35 +140,22 @@ class ScopedInjectable[R, T](Injectable[T], ABC):
         raise NotImplementedError
 
     async def aget_instance(self) -> T:
-        scope = self.get_scope()
-
-        with suppress(KeyError):
-            return scope.cache[self]
-
-        instance = await self.abuild(scope)
-        self.set_instance(instance, scope)
-        return instance
+        scope = self.__get_scope()
+        factory = partial(self.abuild, scope)
+        return await self.logic.aget_or_create(scope.cache, self, factory)
 
     def get_instance(self) -> T:
-        scope = self.get_scope()
-
-        with suppress(KeyError):
-            return scope.cache[self]
-
-        instance = self.build(scope)
-        self.set_instance(instance, scope)
-        return instance
-
-    def get_scope(self) -> Scope:
-        return get_scope(self.scope_name)
-
-    def set_instance(self, instance: T, scope: Scope) -> None:
-        scope.cache[self] = instance
+        scope = self.__get_scope()
+        factory = partial(self.build, scope)
+        return self.logic.get_or_create(scope.cache, self, factory)
 
     def unlock(self) -> None:
         if self.is_locked:
             raise RuntimeError(f"To unlock, close the `{self.scope_name}` scope.")
 
+    def __get_scope(self) -> Scope:
+        return get_scope(self.scope_name)
+
 
 class AsyncCMScopedInjectable[T](ScopedInjectable[AsyncContextManager[T], T]):
     __slots__ = ()
@@ -170,11 +190,38 @@ class SimpleScopedInjectable[T](ScopedInjectable[T, T]):
         return self.factory.call()
 
     def unlock(self) -> None:
-        for scope in get_active_scopes(self.scope_name):
-            scope.cache.pop(self, None)
+        remove_scoped_values(self, self.scope_name)
 
 
-@dataclass(repr=False, frozen=True, slots=True)
+@dataclass(repr=False, eq=False, frozen=True, slots=True)
+class ScopedSlotInjectable[T](Injectable[T]):
+    cls: type[T]
+    scope_name: str
+    key: SlotKey[T] = field(default_factory=SlotKey)
+
+    @property
+    def is_locked(self) -> bool:
+        return in_scope_cache(self.key, self.scope_name)
+
+    async def aget_instance(self) -> T:
+        return self.get_instance()
+
+    def get_instance(self) -> T:
+        scope_name = self.scope_name
+        scope = get_scope(scope_name)
+
+        try:
+            return scope.cache[self.key]
+        except KeyError as exc:
+            raise EmptySlotError(
+                f"The slot for `{self.cls}` isn't set in the current `{scope_name}` scope."
+            ) from exc
+
+    def unlock(self) -> None:
+        remove_scoped_values(self.key, self.scope_name)
+
+
+@dataclass(repr=False, eq=False, frozen=True, slots=True)
 class ShouldBeInjectable[T](Injectable[T]):
     cls: type[T]
 
